'use strict';

const config = require('./../../config');
const logger = require('./../../utils/logger');
const errors = require('./../../utils/dz-errors-api');
const dbConstants = require('./../../constants/db-constants');
const query = require('./../../utils/query-creator-api');
const passwordHandler = require('./../../utils/password');
let async = require('async');
let _ = require('underscore');
const AWS = require('aws-sdk');
AWS.config.update({
    accessKeyId: config.aws.keyId,
    secretAccessKey: config.aws.key,
    region: config.aws.sesRegion
});
const ses = new AWS.SES({ apiVersion: '2010-12-01' });
const commonHandler = require('./../../model_handlers/api/common-handler');
const emailHandler = require('./../../model_handlers/api/email-handler');
const { sendSMS } = require('./../../utils/sms-manager');
const Verify_mobile = require('./../../models/verify-mobile');
const Sms_template = require('./../../models/sms-template');
const Customer = require('./../../models/customer');
const Property = require('./../../models/property');
const seeker_favourite = require('./../../models/seeker-favourite');
const reservation = require('./../../models/reservation');
const pricing = require('./../../models/pricing');
const property_ratings = require('./../../models/property-rating');
const owner_ratings = require('./../../models/owner-rating');
const seeker_ratings = require('./../../models/seeker-rating');
const payouts = require('./../../models/payouts');
const LD = require('lodash');

const setupEmail = async (requestParam) => {
    const params = {
        Destination: {
            ToAddresses: requestParam.to_email
        },
        Message: {
            Body: {
                Html: {
                    Charset: 'UTF-8',
                    Data: requestParam.description
                }
            },
            Subject: {
                Charset: 'UTF-8',
                Data: requestParam.subject
            }
        },
        ReturnPath: requestParam.from_email,
        Source: requestParam.from_email,
    };

    ses.sendEmail(params, (err, data) => {
        if (err) {
            return console.log(err, err.stack);
        } else {
            console.log("Email sent.", data);
        }
    });
};

/*
    Name : Check social id
    Purpose : Check social id
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const checkSocialIdExist = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.countRecord(dbConstants.dbSchema.customers, { social_id: requestParam.social_id });
            if (response == 0) {
                reject(errors.socialIdNotFound(true, requestParam.code));
                return;
            }
            resolve(profile({ social_id: requestParam.social_id }, requestParam.code));
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : Sign In
    Purpose : Sign In
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const signIn = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let columnAndValues = {}
            columnAndValues.mobile = requestParam.mobile;
            columnAndValues.mobile_country_code = requestParam.mobile_country_code;

            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, columnAndValues, { _id: 0, customer_id: 1, password: 1, status: 1 });
            if (!response) {
                reject(errors.mobileNotFound(true, requestParam.code));
                return;
            }

            let encryptPassword = await passwordHandler.encrypt(requestParam.password.toString());
            if (encryptPassword != response.password) {
                reject(errors.invalidPassword(true, requestParam.code));
                return;
            }

            if (response.status == 'inactive') {
                reject(errors.notActivate(true, requestParam.code));
                return;
            }
            let updateColumn = {
                last_login: new Date()
            }
            if (requestParam.device_token && requestParam.device_id && requestParam.device_type) {
                updateColumn.device_token = requestParam.device_token
                updateColumn.device_type = requestParam.device_type
                updateColumn.device_id = requestParam.device_id
            }
            await query.updateSingle(dbConstants.dbSchema.customers, updateColumn, columnAndValues);
            config.firebase.firebaseDBRef.ref('users/' + response.customer_id + '/').set({status:'active'});
            resolve(profile({ customer_id: response.customer_id }, requestParam.code));
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};


/*
    Name : Sign up customer
    Purpose : Sign up customer
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const signUp = async (requestParam, req) => {
    return new Promise(async (resolve, reject) => {
        try {
            if (!requestParam.password) {
                requestParam.password = ''
            }
            let columnAndValues = {}, insertColumnAndValues = { register_type: requestParam.register_type || 'mobile', name: requestParam.name || '', email: requestParam.email || '', mobile_country_code: requestParam.mobile_country_code || '', mobile: requestParam.mobile || '' };

            if (requestParam.register_type == 'mobile') {
                columnAndValues = {
                    mobile: requestParam.mobile,
                    mobile_country_code: requestParam.mobile_country_code
                }
                insertColumnAndValues.password = await passwordHandler.encrypt(requestParam.password.toString())
            } else if (requestParam.register_type == 'google' || requestParam.register_type == 'apple') {
                columnAndValues['$or'] = [{
                    social_id: requestParam.social_id
                }]
                if (requestParam.email && requestParam.email != '') {
                    columnAndValues['$or'].push({
                        email: (new RegExp(['^', requestParam.email.trim(), '$'].join(''), 'i'))
                    });
                }
                if (requestParam.mobile_country_code && requestParam.mobile) {
                    columnAndValues['$or'].push({
                        mobile: requestParam.mobile,
                        mobile_country_code: requestParam.mobile_country_code
                    });
                }

                insertColumnAndValues['social_id'] = requestParam.social_id;
                insertColumnAndValues['profile_picture'] = requestParam.profile_picture;
            }

            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, columnAndValues, { customer_id: 1, social_id: 1, register_type: 1, name: 1, mobile_country_code: 1, mobile: 1 });
            if (response) {
                if (requestParam.email && requestParam.email != '') {
                    if (response.email == requestParam.email) {
                        reject(errors.emailAlreadyExist(true, requestParam.code));
                        return;
                    }
                }
                if (requestParam.mobile_country_code && requestParam.mobile) {
                    if (response.mobile_country_code == requestParam.mobile_country_code && response.mobile == requestParam.mobile) {
                        reject(errors.mobileAlreadyExist(true, requestParam.code));
                        return;
                    }
                }
                if (response.social_id == requestParam.social_id) {
                    reject(errors.socialIdAlreadyExist(true, requestParam.code));
                    return;
                }
            }
            if (req.files) {
                if (req.files.profile_picture) {
                    insertColumnAndValues.profile_picture = await new Promise((solve, reject) => {
                        commonHandler.uploadImage(req.files.profile_picture, config.aws.s3.customerBucket, (error, path) => {
                            solve(path)
                        });
                    });
                }
            }
            if (requestParam.device_token && requestParam.device_id && requestParam.device_type) {
                insertColumnAndValues.device_token = requestParam.device_token;
                insertColumnAndValues.device_id = requestParam.device_id;
                insertColumnAndValues.device_type = requestParam.device_type;
            }

            let language = await query.selectWithAndOne(dbConstants.dbSchema.languages, { code: 'EN' }, { _id: 0, language_id: 1, title: 1 });
            insertColumnAndValues.language_id = language ? language.language_id : ''


            let measurement = await query.selectWithAndOne(dbConstants.dbSchema.measurements, {}, { _id: 0, measurement_id: 1, title: 1 });
            insertColumnAndValues.measurement_id = measurement ? measurement.measurement_id : ''

            let userInfo = await query.insertSingle(dbConstants.dbSchema.customers, insertColumnAndValues);
            // requestParam.template_code = 'CUS_SIGN_UP';
            // emailHandler.sendEmail(requestParam);
            config.firebase.firebaseDBRef.ref('users/' + userInfo.customer_id + '/').set({status:'active'});
            resolve(profile({ customer_id: userInfo.customer_id }, requestParam.code));
            return;
        } catch (error) {
            console.log(error);
            reject(error)
            return
        }
    })
};

/*
    Name : Check Email Mobile customer
    Purpose : Check Email Mobile customer
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const checkEmailMobileExists = async (requestParam, done) => {
    return new Promise(async (resolve, reject) => {
        try {
            let columnAndValues;
            columnAndValues = {
                mobile: requestParam.mobile,
                mobile_country_code: requestParam.mobile_country_code
            }
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, columnAndValues, { social_id: 1, register_type: 1, name: 1, mobile_country_code: 1, mobile: 1 });
            if (response) {
                if (requestParam.mobile_country_code && requestParam.mobile) {
                    if (response.mobile_country_code == requestParam.mobile_country_code && response.mobile == requestParam.mobile) {
                        reject(errors.mobileAlreadyExist(true, requestParam.code));
                        return;
                    }
                }
            }
            else {
                let otp = Math.floor(1000 + Math.random() * 9000);
                let template = await query.selectWithAndOne(dbConstants.dbSchema.sms_templates, { code: 'OTP' }, { _id: 0, value: 1 });
                if (template) {
                    let msg = template.value[requestParam.code]
                    msg = msg.replace('#NAME#', requestParam.name)
                    msg = msg.replace('#OTP#', otp)
                    sendSMS({ message: msg, mobile_country_code: requestParam.mobile_country_code, mobile: requestParam.mobile });
                }
                requestParam.otp = otp;
                let verify = await query.insertSingle(dbConstants.dbSchema.verify_mobiles, requestParam);
                resolve({ verify_mobile_id: verify.verify_mobile_id, otp: otp });
                return
            }
        } catch (error) {
            console.log(error)
            reject(errors.internalServer(true, requestParam.code));
            return;
        }
    })
};

/*
    Name : Verify Mobile
    Purpose : Verify Mobile
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const verifyMobile = async (requestParam, done) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.verify_mobiles, { verify_mobile_id: requestParam.verify_mobile_id }, { verify_mobile_id: 1, otp: 1 });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            else {
                if (response.otp == requestParam.otp) {
                    await query.removeMultiple(dbConstants.dbSchema.verify_mobiles, { verify_mobile_id: requestParam.verify_mobile_id })
                    resolve({})
                    return;
                }
                else {
                    reject(errors.invalidOTP(true, requestParam.code));
                    return;
                }
            }
        } catch (error) {
            console.log(error)
            reject(errors.internalServer(true, requestParam.code));
            return;
        }
    })
};

/*
    Name : Resend OTP
    Purpose : Resend OTP
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const resendOtpUsingMobileVerify = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.verify_mobiles, { verify_mobile_id: requestParam.verify_mobile_id }, { _id: 0, mobile_country_code: 1, mobile: 1 });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let otp = Math.floor(1000 + Math.random() * 9000);
            let template = await query.selectWithAndOne(dbConstants.dbSchema.sms_templates, { code: 'OTP' }, { _id: 0, value: 1 });
            if (template) {
                let msg = template.value[requestParam.code]
                msg = msg.replace('#NAME#', requestParam.name)
                msg = msg.replace('#OTP#', otp)
                sendSMS({ message: msg, mobile_country_code: response.mobile_country_code, mobile: response.mobile });
            }
            await query.updateSingle(dbConstants.dbSchema.verify_mobiles, { otp: otp }, { verify_mobile_id: requestParam.verify_mobile_id });
            resolve({ otp: otp });
            return
        } catch (error) {
            reject(error)
            return
        }
    })
};

/*
    Name : Get profile
    Purpose : Get profile
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const profile = async (columnAndValues, code) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, columnAndValues, { _id: 0, customer_id: 1, language_id: 1, measurement_id: 1, name: 1, surname: 1, mobile_country_code: 1, mobile: 1, status: 1, profile_picture: 1, invitation_code: 1, social_id: 1, reservation_updates: 1, payment_updates: 1, message_updates: 1, general_updates: 1, bank_details: 1, user_type: 1, bank_details_added: 1 });
            if (!response) {
                reject(errors.userNotFound(true, code));
                return;
            }

            if (response.status == 'inactive') {
                reject(errors.notActivate(true, code));
                return;
            }
            response.bank_details.payment_frequency_id =(response.bank_details.payment_frequency_id == null )? "" : response.bank_details.payment_frequency_id;
            response = JSON.parse(JSON.stringify(response))

            response.language_name = ''
            if (response.language_id != '') {
                let language = await query.selectWithAndOne(dbConstants.dbSchema.languages, { language_id: response.language_id }, { _id: 0, language_id: 1, title: 1 });
                response.language_name = language ? language.title : ''
            }

            response.measurement_name = ''
            response.measurement_unit = ''
            if (response.measurement_id != '') {
                let measurement = await query.selectWithAndOne(dbConstants.dbSchema.measurements, { measurement_id: response.measurement_id }, { _id: 0, measurement_id: 1, title: 1, unit: 1 });
                response.measurement_name = measurement ? measurement.title[code] : ''
                response.measurement_unit = measurement ? measurement.unit : ''
            }

            let owner_rate = await query.selectWithAnd(dbConstants.dbSchema.owner_ratings, { owner_id: response.customer_id }, { _id: 0, rating: 1 });
            let rating = 0;
            if (owner_rate.length > 0) {
                rating = LD.sumBy(owner_rate, 'rating')
                rating = parseFloat(rating) / owner_rate.length;
            }
            response.owner_rating = parseFloat(rating).toFixed(1);

            let seeker_rate = await query.selectWithAnd(dbConstants.dbSchema.seeker_ratings, { seeker_id: response.customer_id }, { _id: 0, rating: 1 });
            let ratingSeeker = 0;
            if (seeker_rate.length > 0) {
                ratingSeeker = LD.sumBy(seeker_rate, 'rating')
                ratingSeeker = parseFloat(ratingSeeker) / seeker_rate.length;
            }
            response.seeker_rating = parseFloat(ratingSeeker).toFixed(1);

            if (response.profile_picture != '') {
                response.profile_picture = config.aws.prefix + config.aws.s3.customerBucket + '/' + response.profile_picture
            }
            resolve(response)
            return
        } catch (error) {
            console.log(error);
            reject(error)
            return
        }
    })
};

/*
    Name : Forgot Password
    Purpose : Forgot Password
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const forgotPassword = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let columnAndValues = {}
            columnAndValues.mobile = requestParam.mobile;
            columnAndValues.mobile_country_code = requestParam.mobile_country_code;
            let db = dbConstants.dbSchema.customers;
            let response = await query.countRecord(db, columnAndValues);
            if (response == 0) {
                reject(errors.mobileNotFound(true, requestParam.code));
                return;
            }
            resolve(sendEmailOtp(requestParam, columnAndValues, db));
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : Send Email or OTP
    Purpose : Send Email or OTP
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const sendEmailOtp = async (requestParam, columnAndValues, db) => {
    return new Promise(async (resolve, reject) => {
        try {
            let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, { _id: 0 });
            let response = await query.selectWithAndOne(db, columnAndValues, { _id: 0, name: 1 });
            let otp = Math.floor(1000 + Math.random() * 9000);
            await query.updateSingle(db, { otp: otp }, columnAndValues);
            let template = await query.selectWithAndOne(dbConstants.dbSchema.sms_templates, { code: 'OTP' }, { _id: 0, value: 1 });
            if (template) {
                let msg = template.value[requestParam.code]
                msg = msg.replace('#NAME#', response.name)
                msg = msg.replace('#OTP#', otp)
                sendSMS({ message: msg, mobile_country_code: requestParam.mobile_country_code, mobile: requestParam.mobile });
            }
            resolve({ otp: otp })
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : Verify OTP
    Purpose : Verify OTP
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const verifyOTP = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let columnAndValues = {}
            columnAndValues.mobile = requestParam.mobile;
            columnAndValues.mobile_country_code = requestParam.mobile_country_code;
            let db = dbConstants.dbSchema.customers;
            let response = await query.selectWithAndOne(db, columnAndValues, { _id: 0, otp: 1 });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            if (response.otp == requestParam.otp) {
                await query.updateSingle(db, { otp: 0 }, columnAndValues);
                resolve({})
                return;
            }
            else {
                reject(errors.invalidOTP(true, requestParam.code));
                return;
            }
        } catch (error) {
            reject(error)
            return
        }
    })
};

/*
    Name : Reset Pwd
    Purpose : Reset Pwd
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const resetPassword = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let columnAndValues = {}
            columnAndValues.mobile = requestParam.mobile;
            columnAndValues.mobile_country_code = requestParam.mobile_country_code;
            let db = dbConstants.dbSchema.customers;
            let response = await query.selectWithAndOne(db, columnAndValues, { _id: 0, otp: 1 });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let password = await passwordHandler.encrypt(requestParam.password.toString())
            await query.updateSingle(db, { password: password }, columnAndValues);
            resolve({});
        } catch (error) {
            reject(error)
            return
        }
    })
};

/*
    Name : Logout Delete Account
    Purpose : Logout Delete Account
    Original Author : Gaurav Patel
    Created At : 24th feb 2021
*/
const logoutDeleteAccount = async (requestParam,) => {
    return new Promise(async (resolve, reject) => {
        try {
            let db = dbConstants.dbSchema.customers;
            let getColumn = {
                customer_id: requestParam.customer_id
            }
            let removeColumn = {
                customer_id: { $in: [requestParam.customer_id] }
            }
            let response = await query.selectWithAndOne(db, getColumn, { _id: 0, name: 1 });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            if (requestParam.action == 'logout') {
                await query.updateSingle(db, { player_id: '', device_token: '' }, getColumn);
                resolve({});
                return
            }
            else {
                await query.removeMultiple(db, removeColumn);
                resolve({});
                return
            }
        } catch (error) {
            reject(error)
            return
        }
    })
};

/*
    Name : Update profile
    Purpose : Update profile
    Original Author : Gaurav Patel
    Created At : 5th Mar 2021
*/
const updateProfile = async (requestParam, req) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, { customer_id: requestParam.customer_id }, { _id: 0, customer_id: 1, email: 1, profile_picture: 1, name: 1, mobile: 1, bank_details_added: 1, bank_details: 1 });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            if (requestParam.email) {
                requestParam.email = requestParam.email.trim();
                let regexEmail = new RegExp(['^', requestParam.email, '$'].join(''), 'i');
                let compareColumnAndValues = {
                    $and: [{
                        $or: [{
                            email: regexEmail
                        }]
                    }, {
                        customer_id: {
                            $ne: requestParam.customer_id
                        }
                    }]

                };
                let exists = await query.selectWithAndOne(dbConstants.dbSchema.customers, compareColumnAndValues, { _id: 0, customer_id: 1, email: 1, profile_picture: 1 });
                if (exists) {
                    reject(errors.emailAlreadyExist(true, requestParam.code));
                    return;
                }
            }
            if (requestParam.mobile_country_code && requestParam.mobile) {
                let compareColumnAndValues = {
                    $and: [{
                        $or: [{
                            mobile: requestParam.mobile,
                            mobile_country_code: requestParam.mobile_country_code,
                        }]
                    }, {
                        customer_id: {
                            $ne: requestParam.customer_id
                        }
                    }]

                };
                let exists = await query.selectWithAndOne(dbConstants.dbSchema.customers, compareColumnAndValues, { _id: 0, customer_id: 1, email: 1, profile_picture: 1 });
                if (exists) {
                    reject(errors.mobileAlreadyExist(true, requestParam.code));
                    return;
                }
            }
            if (req.files) {
                if (req.files.profile_picture) {
                    let fileObjects = [{
                        Key: 'parksie/customers/' + /[^/]*$/.exec(response.profile_picture)[0]
                    }];
                    requestParam.profile_picture = await new Promise((solve, reject) => {
                        commonHandler.removeMultipleImages(fileObjects, (error, path) => {
                            commonHandler.uploadImage(req.files.profile_picture, config.aws.s3.customerBucket, (error, path) => {
                                solve(path)
                            });
                        });
                    });
                }
            }
            if (requestParam.payment_frequency_id) {
                let bank_details = response.bank_details
                bank_details.payment_frequency_id = requestParam.payment_frequency_id || ""
                requestParam.bank_details = bank_details
            }
            await query.updateSingle(dbConstants.dbSchema.customers, requestParam, { customer_id: requestParam.customer_id });
            resolve(profile({ customer_id: requestParam.customer_id }, requestParam.code));
            return
        } catch (error) {
            reject(error)
            return
        }
    })
};

/*
    Name : Update notification update
    Purpose : Update notification update
    Original Author : Gaurav Patel
    Created At : 8th Mar 2021
*/
const updateNotification = async (requestParam, req) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, { customer_id: requestParam.customer_id }, { _id: 0, customer_id: 1, email: 1, profile_picture: 1, name: 1, mobile: 1, reservation_updates: 1, payment_updates: 1, message_updates: 1, general_updates: 1 });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let obj = {}
            obj[requestParam.type] = {
                sms: requestParam.action == 'sms' ? (requestParam.value == 'true' ? true : false) : response[requestParam.type]['sms'],
                in_app: requestParam.action == 'in_app' ? (requestParam.value == 'true' ? true : false) : response[requestParam.type]['in_app'],
            }
            await query.updateSingle(dbConstants.dbSchema.customers, obj, { customer_id: requestParam.customer_id });
            resolve(profile({ customer_id: requestParam.customer_id }, requestParam.code));
            return
        } catch (error) {
            reject(error)
            return
        }
    })
};

/*
    Name : Change Password
    Purpose : Change Password
    Original Author : Gaurav Patel
    Created At : 8th Mar 2021
*/
const changePassword = async (requestParam, req) => {
    return new Promise(async (resolve, reject) => {
        try {
            let db = dbConstants.dbSchema.customers;
            let getColumn = {
                customer_id: requestParam.customer_id
            }
            let response = await query.selectWithAndOne(db, getColumn, { _id: 0, password: 1 });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let encryptPassword = await passwordHandler.encrypt(requestParam.old_password.toString());
            if (encryptPassword != response.password) {
                reject(errors.invalidOldPassword(true, requestParam.code));
                return;
            }
            let password = await passwordHandler.encrypt(requestParam.new_password.toString());
            await query.updateSingle(db, { password: password }, getColumn);
            resolve({});
            return
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};


module.exports = {
    checkSocialIdExist,
    profile,
    signIn,
    signUp,
    checkEmailMobileExists,
    verifyMobile,
    resendOtpUsingMobileVerify,
    forgotPassword,
    verifyOTP,
    resetPassword,
    logoutDeleteAccount,
    updateProfile,
    changePassword,
    updateNotification
};