'use strict';

const config = require('./../../config');
const logger = require('./../../utils/logger');
const errors = require('./../../utils/dz-errors-api');
const dbConstants = require('./../../constants/db-constants');
const query = require('./../../utils/query-creator-api');
let async = require('async');
let _ = require('underscore');
const Chat = require('./../../models/chat');
const commonHandler = require('./../../model_handlers/api/common-handler');
const notificationHandler = require('./../../model_handlers/api/notification-handler');

/*
    Name : Add chat
    Purpose : Add chat
    Original Author : Gaurav Patel
    Created At : 1st March 2021
*/
const add = async(requestParam) => {
    return new Promise(async(resolve, reject) => {
        try {
            requestParam.code = 'EN';
            let res = await query.selectWithAndOne(dbConstants.dbSchema.chats, { room: requestParam.room }, { _id: 0}, { created_at: 1 });
            let msg  = requestParam.message;
            if(requestParam.message_type == 'image'){
                requestParam.message = /[^/]*$/.exec(requestParam.message)[0]
            }
            let reservation = await query.selectWithAndOne(dbConstants.dbSchema.reservations, {reservation_id: requestParam.room}, { _id: 0, property_id: 1} );
            let address = ''
            if(reservation){
                let property = await query.selectWithAndOne(dbConstants.dbSchema.properties, {property_id: reservation.property_id}, { _id: 0, address:1} );
                address = property ? property.address : ''
            }
            let profile_user = await query.selectWithAndOne(dbConstants.dbSchema.customers, {customer_id: requestParam.user_id}, { _id: 0, name:1, device_token:1,profile_picture:1 }, { created_at: 1 });
            let user = await query.selectWithAndOne(dbConstants.dbSchema.customers, {customer_id: requestParam.opponent_user_id}, { _id: 0, name:1, device_token:1,profile_picture:1 }, { created_at: 1 });
            if(user){
                notificationHandler.sendNotification({
                    message:msg,
                    type:'chat_msg',
                    title:config.appName,
                    user_id: requestParam.opponent_user_id, 
                    sender_id: requestParam.user_id, 
                    reservation_id: requestParam.room,
                    user_type:'', 
                    reservation_status:'', 
                    device_token:[user.device_token],
                });
                notificationHandler.setLogs({
                    user_id:[requestParam.opponent_user_id],
                    user_type:'',
                    message: msg, 
                    description:msg,
                    type:'chat_msg',
                    profile_picture:profile_user ? profile_user.profile_picture : '',
                    name:profile_user ? profile_user.name : '',
                    address: address,
                    sender_id: requestParam.user_id,
                    reservation_id: requestParam.room,
                    reservation_status: ''
                });
            }
            if(res){
                let resInfo = await query.updateSingle(dbConstants.dbSchema.chats, {
                    $push: {
                        chat: {
                            message: requestParam.message,
                            user_id: requestParam.user_id,
                            opponent_user_id: requestParam.opponent_user_id,
                            message_type: requestParam.message_type,
                            datetime: requestParam.datetime,
                            status: []
                        }
                    }
                }, { room: requestParam.room });
            }
            else{
                let insertObj = {
                room: requestParam.room,
                    chat: [{
                        message: requestParam.message,
                        user_id: requestParam.user_id,
                        opponent_user_id: requestParam.opponent_user_id,
                        message_type: requestParam.message_type,
                        datetime: requestParam.datetime,
                        status: []
                    }]
                }
                let resInfo = await query.insertSingle(dbConstants.dbSchema.chats, insertObj);
            }
            return false;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : List chat
    Purpose : List chat
    Original Author : Gaurav Patel
    Created At : 1st March 2021
*/
const list = async(requestParam) => {
    return new Promise(async(resolve, reject) => {
        try {
            let res = await query.selectWithAndOne(dbConstants.dbSchema.chats, { room: requestParam.room }, { _id: 0}, { created_at: 1 });
            if(!res){
                resolve({});
                return;
            }
            else{
                let readArr = [];
                _.each(res.chat, (elem) => {
                    let status = elem.status;
                    if(status.includes(requestParam.user_id) == false){
                        status.push(requestParam.user_id)
                    }
                    elem.status = _.flatten(status)
                    readArr.push(elem)
                    if(elem.message_type == 'image'){
                        elem.message = elem.message!='' ? config.aws.prefix + config.aws.s3.chatBucket + '/' + elem.message : ''
                    }
                })

                let resInfo = await query.updateSingle(dbConstants.dbSchema.chats, {
                    chat:readArr
                }, { room: requestParam.room });

                resolve(res);
                return;
            }
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : read chat msg
    Purpose : read chat msg
    Original Author : Gaurav Patel
    Created At : 8th June 2021
*/
const readMsg = async(requestParam) => {
    return new Promise(async(resolve, reject) => {
        try {
            let res = await query.selectWithAndOne(dbConstants.dbSchema.chats, { room: requestParam.room }, { _id: 0}, { created_at: 1 });
            if(!res){
                resolve({});
                return;
            }
            else{
                let readArr = [];
                _.each(res.chat, (elem) => {
                    let status = elem.status;
                    if(status.includes(requestParam.user_id) == false){
                        status.push(requestParam.user_id)
                    }
                    elem.status = _.flatten(status)
                    readArr.push(elem)
                })
                
                let resInfo = await query.updateSingle(dbConstants.dbSchema.chats, {
                    chat:readArr
                }, { room: requestParam.room });

                resolve({});
                return;
            }
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : Upload or Delete Package Image
    Purpose : Upload or Delete Package Image
    Original Author : Gaurav Patel
    Created At : 2nd Mar 2021
*/
const uploadImage = async(requestParam, req) => {
    return new Promise(async(resolve, reject) => {
        try {
            let photo = await new Promise((solve, reject) => {
                commonHandler.uploadImage(req.files.photo, config.aws.s3.chatBucket, (error, path) => {
                    solve(path)
                });
            });
            resolve(config.aws.prefix + config.aws.s3.chatBucket + '/' + photo);
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

module.exports = {
    add,
    list,
    uploadImage,
    readMsg
};