'use strict';

const config = require('./../../config');
const logger = require('./../../utils/logger');
const errors = require('./../../utils/dz-errors-api');
const dbConstants = require('./../../constants/db-constants');
const query = require('./../../utils/query-creator-api');
let async = require('async');
let _ = require('underscore');
const vehicle = require('./../../models/vehicle');
const commonHandler = require('./../../model_handlers/api/common-handler');

/*
	Name : List
	Purpose : List
	Original Author : Gaurav Patel
    Created At : 8th Mar 2021
*/
const list = async(requestParam) => {
    return new Promise(async(resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {customer_id: requestParam.customer_id}, { _id: 0, customer_id: 1} );
            if(!response){
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let joinArr = [{
                $lookup: {
                    from: 'vehicle_types',
                    localField: 'vehicle_type_id',
                    foreignField: 'vehicle_type_id',
                    as: 'typeDetails'
                }
            },  {
                $unwind: "$typeDetails"
            },  {
                $lookup: {
                    from: 'vehicle_colors',
                    localField: 'vehicle_color_id',
                    foreignField: 'vehicle_color_id',
                    as: 'colorDetails'
                }
            },  {
                $unwind: "$colorDetails"
            },  { 
                $match : { customer_id: requestParam.customer_id }
            }, { 
                $sort : {created_at:-1}
            }, {
                $project: {
                    _id: 0,
                    customer_id: "$customer_id",
                    vehicle_type_id: "$vehicle_type_id",
                    vehicle_color_id: "$vehicle_color_id",
                    vehicle_id: "$vehicle_id",
                    photo: "$photo",
                    vehicle_number: "$vehicle_number",
                    vehicle_color: "$colorDetails.title."+requestParam.code,
                    vehicle_type: "$typeDetails.title."+requestParam.code,
                    is_default: "$is_default",
                    is_electric: "$is_electric",
                    vehicle_make: "$vehicle_make",
                    vehicle_model: "$vehicle_model",
                }
            }];
            let vehicles = await query.joinWithAnd(dbConstants.dbSchema.vehicles, joinArr);
            _.each(vehicles, (elem) => {
                elem.photo = elem.photo !='' ? config.aws.prefix + config.aws.s3.vehicleBucket + '/' + elem.photo : ''
            });
            resolve(vehicles)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : add
    Purpose : add
    Original Author : Gaurav Patel
    Created At : 8th Mar 2021
*/
const add = async(requestParam, req) => {
    return new Promise(async(resolve, reject) => {
        console.log("adding vehicle api called here")
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {customer_id: requestParam.customer_id}, { _id: 0, customer_id: 1} );
            if(!response){
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            if(requestParam.type == 'add'){
                let added_vehicle = await query.selectWithAndOne(dbConstants.dbSchema.vehicles, {vehicle_type_id: requestParam.vehicle_type_id, customer_id: requestParam.customer_id, vehicle_number:requestParam.vehicle_number}, { _id: 0, vehicle_id: 1} );
                if(added_vehicle){
                    reject(errors.vehicleAlreadyAdded(true, requestParam.code));
                    return;
                }
            }
            if(req.files){
                if(req.files.photo){
                    requestParam.photo = await new Promise((solve, reject) => {
                        commonHandler.uploadImage(req.files.photo, config.aws.s3.vehicleBucket, (error, path) => {
                            solve(path)
                        });
                    });
                }
            }
            let vehicles = await query.selectWithAnd(dbConstants.dbSchema.vehicles, {customer_id: requestParam.customer_id}, { _id: 0, vehicle_id: 1} );
            if(vehicles.length == 0){
                requestParam.is_default = true
            }
            if(requestParam.is_default == true || requestParam.is_default == 'true'){
                await query.updateMultiple(dbConstants.dbSchema.vehicles, {is_default:false}, {customer_id: requestParam.customer_id});
            }
            console.log("Nikunj Req Params",requestParam)
            if(requestParam.type == 'add'){
                if(requestParam.is_electric == false || requestParam.is_electric == 'false' || requestParam.is_electric == ''){

                    requestParam.is_electric = false
                    console.log("requestParam.is_electric>>>here",requestParam.is_electric)
                }else{
                    requestParam.is_electric = true
                }
                console.log("requestParam.is_electric>>>",requestParam.is_electric)
                await query.insertSingle(dbConstants.dbSchema.vehicles, requestParam);
            }
            else{
                await query.updateSingle(dbConstants.dbSchema.vehicles, requestParam, {vehicle_id: requestParam.vehicle_id});
            }
            resolve(list({ customer_id: requestParam.customer_id, code: requestParam.code }))
            return;
        } catch (error) {
            reject(error)
            return
        }
    })
};

/*
    Name : Set Default
    Purpose : Set Default
    Original Author : Gaurav Patel
    Created At : 12th March 2021
*/
const setDefault = async(requestParam, req) => {
    return new Promise(async(resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {customer_id: requestParam.customer_id}, { _id: 0, customer_id: 1} );
            if(!response){
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let vehicle = await query.selectWithAndOne(dbConstants.dbSchema.vehicles, {vehicle_id: requestParam.vehicle_id}, { _id: 0, vehicle_id: 1} );
            if(!vehicle){
                reject(errors.vehicleNotFound(true, requestParam.code));
                return;
            }
            await query.updateMultiple(dbConstants.dbSchema.vehicles, {is_default:false}, {customer_id: requestParam.customer_id});
            await query.updateSingle(dbConstants.dbSchema.vehicles, { is_default: true }, {vehicle_id: requestParam.vehicle_id});
            resolve(list({ customer_id: requestParam.customer_id, code: requestParam.code }))
            return;
        } catch (error) {
            reject(error)
            return
        }
    })
};

/*
    Name : Set Default
    Purpose : Set Default
    Original Author : Gaurav Patel
    Created At : 12th March 2021
*/
const deleteVehicle = async(requestParam, req) => {
    return new Promise(async(resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {customer_id: requestParam.customer_id}, { _id: 0, customer_id: 1} );
            if(!response){
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let vehicle = await query.selectWithAndOne(dbConstants.dbSchema.vehicles, {vehicle_id: requestParam.vehicle_id}, { _id: 0, vehicle_id: 1} );
            if(!vehicle){
                reject(errors.vehicleNotFound(true, requestParam.code));
                return;
            }
            await query.removeMultiple(dbConstants.dbSchema.vehicles, {vehicle_id: requestParam.vehicle_id})
            resolve(list({ customer_id: requestParam.customer_id, code: requestParam.code }))
            return;
        } catch (error) {
            reject(error)
            return
        }
    })
};

module.exports = {
    list,
    add,
    setDefault,
    deleteVehicle
};