'use strict';

const logger = require('./../../utils/logger');
const jsonResponse = require('./../../utils/json-response');
const errors = require('./../../utils/dz-errors');
const dbConstants = require('./../../constants/db-constants');
const query = require('./../../utils/query-creator');
let async = require('async');
let _ = require('underscore');
const config = require('./../../config');
let fs = require('fs');
let moment = require('moment');
const timeZone = require('moment-timezone');

const get = function(req,done){
    let columnAndValue = {}
    if(req.query.is_dashboard){
        let date = timeZone(new Date()).tz(req.headers.time_zone).format('YYYY-MM-DD');
        columnAndValue.created_at = {
            $lt: new Date(date+'T23:59:59.000Z'),
            $gte: new Date(date+'T00:00:00.000Z')
        }
    }
    if(req.query.from_date && req.query.to_date){
        if(req.query.from_date != req.query.to_date){
            req.query.to_date = moment(new Date(req.query.to_date)).subtract(1 , 'day');
            req.query.to_date = moment(new Date(req.query.to_date)).format('YYYY-MM-DD')
        }
        columnAndValue.created_at = {
            $lt: new Date(req.query.to_date+'T23:59:59.000Z'),
            $gte: new Date(req.query.from_date+'T00:00:00.000Z')
        }
    }
    if(req.query.type && req.query.value){
        if(req.query.type!='' && req.query.value!=''){
            if(req.query.type == 'customer'){
                columnAndValue.customer_id = req.query.value
            }
            if(req.query.type == 'parking_spot'){
                columnAndValue.property_id = req.query.value
            }
        }
    }
	let joinArr = [{
        $lookup: {
            from: 'properties',
            localField: 'property_id',
            foreignField: 'property_id',
            as: 'propDetails'
        }
    },  {
        $unwind: "$propDetails"
    },  {
        $lookup: {
            from: 'customers',
            localField: 'customer_id',
            foreignField: 'customer_id',
            as: 'cusDetails'
        }
    },  {
        $unwind: "$cusDetails"
    },  {
        $lookup: {
            from: 'customers',
            localField: 'owner_id',
            foreignField: 'customer_id',
            as: 'ownDetails'
        }
    },  {
        $unwind: "$ownDetails"
    },  { 
        $match : columnAndValue
    }, { 
        $sort : {created_at:-1}
    }, {
        $project: {
            _id: 0,
            reservation_id: "$reservation_id",
            title: "$propDetails.address",
            title: "$propDetails.address",
            status: "$status",
            created_at: "$created_at",
            from_date: "$from_date",
            to_date: "$to_date",
            transaction_id: "$transaction_id",
            company_percentage: "$company_percentage",
            owner_earn: "$owner_earn",
            company_earn: "$company_earn",
            price: "$price",
            customer: "$cusDetails.name",
            owner: "$ownDetails.name",
            is_paid: "$is_paid",
            is_refund: "$is_refund",
        }
    }];
    query.joinWithAnd(dbConstants.dbSchema.reservations, joinArr, (error, response) => {
        if (error) {
            logger('Error: can not get record.');
            done(errors.internalServer(true), null);
            return;
        }
        response = JSON.parse(JSON.stringify(response))
        _.each(response, (elem) => {
            elem.is_refund = !elem.is_refund ? false: elem.is_refund
            elem.created_at = timeZone(new Date(elem.created_at)).tz(req.headers.time_zone).format('lll');

            let from_date = timeZone(new Date(elem.from_date)).tz(req.headers.time_zone);
            let to_date = timeZone(new Date(elem.to_date)).tz(req.headers.time_zone)
            let duration = moment.duration(to_date.diff(from_date));
            const hours = parseInt(duration.asHours());
            let lbl;
            if(hours > 1){
                elem.booking_duration = hours
                elem.lbl = ' hrs'
            }
            else{
                const mins = parseInt(duration.asMinutes());
                elem.booking_duration = mins
                elem.lbl = mins > 1 ? ' mins' : ' min'
            }

            elem.from_date = timeZone(new Date(elem.from_date)).tz(req.headers.time_zone).format('lll');
            elem.to_date = timeZone(new Date(elem.to_date)).tz(req.headers.time_zone).format('lll')

            elem.price = parseFloat(elem.price.toFixed(2))
            elem.company_earn = parseFloat(elem.company_earn.toFixed(2))
            elem.owner_earn = parseFloat(elem.owner_earn.toFixed(2))
        });
        done(null, response)
    });
};


const details = function(req,done){
    let joinArr = [{
        $lookup: {
            from: 'properties',
            localField: 'property_id',
            foreignField: 'property_id',
            as: 'propDetails'
        }
    },  {
        $unwind: "$propDetails"
    },  {
        $lookup: {
            from: 'customers',
            localField: 'customer_id',
            foreignField: 'customer_id',
            as: 'cusDetails'
        }
    },  {
        "$unwind": {
            "path": "$cusDetails",
            "preserveNullAndEmptyArrays": true
        }
    }, {
        $lookup: {
            from: 'customers',
            localField: 'owner_id',
            foreignField: 'customer_id',
            as: 'ownerDetails'
        }
    },  {
        "$unwind": {
            "path": "$ownerDetails",
            "preserveNullAndEmptyArrays": true
        }
    }, { 
        $match : {reservation_id: req.query.reservation_id}
    }, { 
        $sort : {created_at:-1}
    }, {
        $project: {
            _id: 0,
            reservation_id: "$reservation_id",
            cusDetails: "$cusDetails",
            ownerDetails: "$ownerDetails",
            address: "$propDetails.address",
            status: "$status",
            created_at: "$created_at",
            from_date: "$from_date",
            to_date: "$to_date",
            price: "$price",
            tax_info: "$tax_info",
            company_earn: "$company_earn",
            owner_earn: "$owner_earn",
            company_percentage: "$company_percentage",
            is_refund: "$is_refund",
        }
    }];
    query.joinWithAnd(dbConstants.dbSchema.reservations, joinArr, (error, response) => {
        if (error) {
            logger('Error: can not get record.');
            done(errors.internalServer(true), null);
            return;
        }
        response = JSON.parse(JSON.stringify(response))
        if(response.length > 0){
            _.each(response, (elem) => {
                elem.is_refund = !elem.is_refund ? false: elem.is_refund
                elem.company_percentage = elem.company_percentage ? elem.company_percentage : 0
                if(!elem.tax_info){
                    elem.tax_info = {title:'',value:0}
                }
                if(!elem.cusDetails){
                    elem.cusDetails = {}
                }
                if(!elem.ownerDetails){
                    elem.ownerDetails = {}
                }
                elem.created_at = timeZone(new Date(elem.created_at)).tz(req.headers.time_zone).format('lll');

                let from_date = timeZone(new Date(elem.from_date)).tz(req.headers.time_zone);
                let to_date = timeZone(new Date(elem.to_date)).tz(req.headers.time_zone)
                let duration = moment.duration(to_date.diff(from_date));
                const hours = parseInt(duration.asHours());
                if(hours > 1){
                    elem.booking_duration = hours+' hrs'
                }
                else{
                    const mins = parseInt(duration.asMinutes());
                    elem.booking_duration = mins > 1 ? mins+' mins' : mins+' min'
                }
                elem.total = parseFloat(elem.tax_info.value) + parseFloat(elem.price).toFixed(2)
                elem.total = parseFloat(elem.total).toFixed(2)
                elem.price = parseFloat(elem.price).toFixed(2)

                elem.from_date = timeZone(new Date(elem.from_date)).tz(req.headers.time_zone).format('lll');
                elem.to_date = timeZone(new Date(elem.to_date)).tz(req.headers.time_zone).format('lll')

                elem.company_earn = elem.company_earn.toFixed(2)
                elem.owner_earn = elem.owner_earn.toFixed(2)
            });
            done(null, response[0])
        }
        else{
            done(null, {cusDetails:{}, ownerDetails:{}, tax_info:{}})
        }
    });
};

const getCustomersSpots = (requestParam, done) =>{
    if(requestParam.type == 'customer'){
        query.selectWithAndFilter(dbConstants.dbSchema.customers, {status:'active', user_type:{$ne:'owner'}}, {
            _id: 0,
            customer_id:1,
            name:1
        }, {created_at:-1}, {}, (error, response) => {
            if (error) {
                logger('Error: can not get record.');
                done(errors.internalServer(true), null);
                return;
            }
            let arr = [];
            _.each(response, (elem) =>{
                arr.push({value:elem.customer_id, title:elem.name})
            })
            done(null, arr)
        });
    }
    else{
        query.selectWithAndFilter(dbConstants.dbSchema.properties, {status:'completed'}, {
            _id: 0,
            address:1,
            property_id:1,
        }, {created_at:-1}, {}, (error, response) => {
            if (error) {
                logger('Error: can not get record.');
                done(errors.internalServer(true), null);
                return;
            }
            let arr = [];
            _.each(response, (elem) =>{
                arr.push({value:elem.property_id, title:elem.address})
            })
            done(null, arr)
        });
    }
}

module.exports = {
	get,
    details,
    getCustomersSpots
};