'use strict';

const logger = require('./../../utils/logger');
const jsonResponse = require('./../../utils/json-response');
const errors = require('./../../utils/dz-errors');
const dbConstants = require('./../../constants/db-constants');
const query = require('./../../utils/query-creator');
const queryApi = require('./../../utils/query-creator-api');
let async = require('async');
let _ = require('underscore');
const moment = require('moment');
const LD = require('lodash');
const monthName = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

const statistics = async function(requestParam,done){
    let matchColumn = {}
    if(requestParam.statistics_val && requestParam.statisticsYear){
        if(requestParam.statistics_val != 'all'){
            let qtr = requestParam.statistics_val.split('-')
            if(qtr.length > 1){
                var firstDay = new Date(parseFloat(requestParam.statisticsYear), parseFloat(qtr[0]), 1);
                firstDay.setHours(0, 0, 0, 0);
                var lastDay = new Date(parseFloat(requestParam.statisticsYear), parseFloat(qtr[1]) + 1, 0);
                lastDay.setHours(23, 59, 59, 999);
            }
            else{
                var firstDay = new Date(parseFloat(requestParam.statisticsYear), parseFloat(requestParam.statistics_val), 1);
                firstDay.setHours(0, 0, 0, 0);
                var lastDay = new Date(parseFloat(requestParam.statisticsYear), parseFloat(requestParam.statistics_val) + 1, 0);
                lastDay.setHours(23, 59, 59, 999);
            }
            matchColumn.created_at = {
                $gte: firstDay,
                $lt: lastDay
            }
        }
        else{
            let firstDay = moment(new Date(requestParam.statisticsYear)).startOf('year').format('YYYY-MM-DD')
            let lastDay = moment(new Date(requestParam.statisticsYear)).endOf('year').format('YYYY-MM-DD')
            matchColumn.created_at = {
                $lte: new Date(lastDay + 'T23:59:59.000Z'),
                $gte: new Date(firstDay + 'T00:00:00.000Z')
            }
        }
    }
    let customers = await queryApi.selectWithAnd(dbConstants.dbSchema.customers, matchColumn, {customer_id: 1,user_type: 1}, {})

    matchColumn.status = 'completed'
    let reservations = await queryApi.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, {reservation_id: 1,price: 1, company_earn:1, status:1}, {})
    
    let revenue = LD.sumBy(reservations, 'price')
    revenue = parseFloat(revenue).toFixed(2)
    let earnings = LD.sumBy(reservations, 'company_earn')
    earnings = parseFloat(earnings).toFixed(2)

    matchColumn.is_owner_paid = false
    let payoutsRes = await queryApi.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, {reservation_id: 1,owner_earn:1}, {})
    let payouts = LD.sumBy(payoutsRes, 'owner_earn')
    payouts = parseFloat(payouts).toFixed(2)

    done(null, {
        owner_users: (_.where(customers, {user_type:'owner'})).length,
        seeker_users: (_.where(customers, {user_type:'seeker'})).length,
        both_users: (_.where(customers, {user_type:'both'})).length,
        revenue:revenue,
        earnings:earnings,
        payouts:payouts,
    });
};

const propertyPtatistics = async function(requestParam,done){
    let matchColumn = {}
    if(requestParam.pro_statistics_val && requestParam.proStatisticsYear){
        if(requestParam.pro_statistics_val != 'all'){
            let qtr = requestParam.pro_statistics_val.split('-')
            if(qtr.length > 1){
                var firstDay = new Date(parseFloat(requestParam.proStatisticsYear), parseFloat(qtr[0]), 1);
                firstDay.setHours(0, 0, 0, 0);
                var lastDay = new Date(parseFloat(requestParam.proStatisticsYear), parseFloat(qtr[1]) + 1, 0);
                lastDay.setHours(23, 59, 59, 999);
            }
            else{
                var firstDay = new Date(parseFloat(requestParam.proStatisticsYear), parseFloat(requestParam.pro_statistics_val), 1);
                firstDay.setHours(0, 0, 0, 0);
                var lastDay = new Date(parseFloat(requestParam.proStatisticsYear), parseFloat(requestParam.pro_statistics_val) + 1, 0);
                lastDay.setHours(23, 59, 59, 999);
            }
            matchColumn.created_at = {
                $gte: firstDay,
                $lt: lastDay
            }
        }
        else{
            let firstDay = moment(new Date(requestParam.proStatisticsYear)).startOf('year').format('YYYY-MM-DD')
            let lastDay = moment(new Date(requestParam.proStatisticsYear)).endOf('year').format('YYYY-MM-DD')
            matchColumn.created_at = {
                $lte: new Date(lastDay + 'T23:59:59.000Z'),
                $gte: new Date(firstDay + 'T00:00:00.000Z')
            }
        }
    }
    let customersAll = await queryApi.selectWithAnd(dbConstants.dbSchema.customers, {}, {customer_id: 1,user_type: 1}, {})
    let properties = await queryApi.countRecord(dbConstants.dbSchema.properties, matchColumn);
    let reservationsCnt = await queryApi.countRecord(dbConstants.dbSchema.reservations, matchColumn);
    done(null, {
        avg_parking_spots_owner: Math.ceil(properties / (_.where(customersAll, {user_type:'owner'})).length),
        avg_reservations_owner: Math.ceil(reservationsCnt / (_.where(customersAll, {user_type:'owner'})).length),
        avg_reservations_spots: Math.ceil(reservationsCnt / properties),
    });
};

const earnings = async(req, done) => {
    let year = new Date().getFullYear()
    if(req.query.year){
        year = parseFloat(req.query.year)
    }
    let columnAndValues = {};
    if(req.query.earning_filter_by == 'month'){
        let promise = [];
        for (let x in monthName) {
            promise.push(await overallEarningSalesCountByMonth(monthName[x], x, columnAndValues, 'company_earn', year))
        }
        Promise.all(promise)
        .then(async result => {
            done(null, result)
        });
    }
    else{
        let promise = [];
        let q = [0, 3, 6, 9];
        for(let x in q){
            promise.push(await quarterlyEarning(q[x], columnAndValues, 'company_earn', year))
        }
        Promise.all(promise)
        .then(async result => {
            let lbl = _.pluck(result, 'label')
            let val = _.pluck(result, 'val')
            done(null, {label:lbl, val:val})
        });
    }
};

const sales = async(req, done) => {
    let year = new Date().getFullYear()
    if(req.query.year){
        year = parseFloat(req.query.year)
    }
    let columnAndValues = {};
    if(req.query.sale_filter_by == 'month'){
        let promise = [];
        for (let x in monthName) {
            promise.push(await overallEarningSalesCountByMonth(monthName[x], x, columnAndValues, 'price', year))
        }
        Promise.all(promise)
        .then(async result => {
            done(null, result)
        });
    }
    else{
        let promise = [];
        let q = [0, 3, 6, 9];
        for(let x in q){
            promise.push(await quarterlyEarning(q[x], columnAndValues, 'price', year))
        }
        Promise.all(promise)
        .then(async result => {
            let lbl = _.pluck(result, 'label')
            let val = _.pluck(result, 'val')
            done(null, {label:lbl, val:val})
        });
    }
};

const overallEarningSalesCountByMonth = (month, index, columnAndValues, column, year) => {
    return new Promise(async(resolve, reject) => {
        var date = new Date(),
            y = year,
            m = parseInt(index);
        var firstDay = new Date(y, m, 1);
        firstDay.setHours(0, 0, 0, 0);
        var lastDay = new Date(y, m + 1, 0);
        lastDay.setHours(23, 59, 59, 999);
        let compairData = {
            status: 'completed',
            completed_date: {
                $gte: firstDay,
                $lt: lastDay
            }
        }
        query.selectWithAnd(dbConstants.dbSchema.reservations, compairData, {
            order_id: 1,
            price: 1,
            company_earn: 1,
        }, (error, reservations) => {
            let monthEarning = LD.sumBy(reservations, column);
            monthEarning = parseFloat(monthEarning).toFixed(2)
            resolve(parseFloat(monthEarning));
            return;
        });
    })
};

const quarterlyEarning = async(index, columnAndValues, column, year) => {
    return new Promise(async(resolve, reject) => {
        try {
            var date = new Date();
            let y = year;
            let m = parseInt(index);
            var firstDay = new Date(y, m, 1);
            firstDay.setHours(0,0,0,0);
            m = m + 2
            var lastDay = new Date(y, m + 1, 0);
            lastDay.setHours(23,59,59,999);
            let compairData = {
                status: 'completed',
                completed_date: {
                    $gte: firstDay,
                    $lt: lastDay
                }
            }
            query.selectWithAnd(dbConstants.dbSchema.reservations, compairData, {
                order_id: 1,
                price: 1,
                company_earn: 1,
            }, (error, reservations) => {
                let monthEarning = LD.sumBy(reservations, column);
                monthEarning = parseFloat(monthEarning).toFixed(2)
                let lbl = 'Jan-Mar';
                if(index == 3){
                    lbl = 'Apr-Jun';
                }
                if(index == 6){
                    lbl = 'Jul-Sep';
                }
                if(index == 9){
                    lbl = 'Oct-Dec';
                }
                resolve({label:lbl, val:parseFloat(monthEarning)});
                return;
            });
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

module.exports = {
	statistics,
    propertyPtatistics,
    sales,
    earnings,
};