'use strict';

let _ = require('underscore');
const express = require('express');
const config = require('./../../config');
const responseCodes = require('./../../helpers/response-codes');
const logger = require('./../../utils/logger');
const jsonResponse = require('./../../utils/json-response');
const errors = require('./../../utils/dz-errors-api');
const router = express.Router();
const seekerHandler = require('./../../model_handlers/api/seeker-handler');

/*
    Name : List
    Purpose : List
    Original Author : Gaurav Patel
    Created At : 18th Mar 2021
*/
router.get('/parking-spots-list', async (req, res) => {
    try {
        req.query.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.query.time_zone = req.headers.time_zone
        }
        if (_.isEmpty(req.query) || !req.query.code || !req.query.customer_id || !req.query.address || !req.query.latitude || !req.query.longitude || !req.query.from_date || !req.query.to_date) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.query.code), null);
            return;
        }

        let response = await seekerHandler.parkingSpotList(req.query);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.query.code), null);
            return;
        }
    }
});

/*
    Name : Favourite List
    Purpose : Favourite List
    Original Author : Gaurav Patel
    Created At : 05th APr 2021
*/
router.get('/favourite-list', async (req, res) => {
    try {
        req.query.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.query.time_zone = req.headers.time_zone
        }
        if (_.isEmpty(req.query) || !req.query.code || !req.query.customer_id) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.query.code), null);
            return;
        }

        let response = await seekerHandler.favouriteList(req.query);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.query.code), null);
            return;
        }
    }
});

/*
    Name : Favourite/Unfavourite
    Purpose : Favourite/Unfavourite
    Original Author : Gaurav Patel
    Created At : 05th APr 2021
*/
router.post('/favourite', async (req, res) => {
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        if (!req.body.code || !req.body.customer_id || !req.body.property_id) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.body.code), null);
            return;
        }

        let response = await seekerHandler.favourite(req.body);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
            return;
        }
    }
});

/*
    Name : reservation
    Purpose : reservation
    Original Author : Gaurav Patel
    Created At : 05th APr 2021
*/
router.post('/reservation', async (req, res) => {
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        if (!req.body.code || !req.body.customer_id || !req.body.property_id || !req.body.vehicle_id || !req.body.card_id || !req.body.driver_details || !req.body.from_date || !req.body.to_date) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.body.code), null);
            return;
        }

        let response = await seekerHandler.reservation(req.body);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
            return;
        }
    }
});

/*
    Name : reservation List
    Purpose : reservation List
    Original Author : Gaurav Patel
    Created At : 05th APr 2021
*/
router.get('/reservation-list', async (req, res) => {
    try {
        req.query.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.query.time_zone = req.headers.time_zone
        }
        if (_.isEmpty(req.query) || !req.query.code || !req.query.customer_id || !req.query.type) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.query.code), null);
            return;
        }

        let response = await seekerHandler.reservationList(req.query);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.query.code), null);
            return;
        }
    }
});

/*
    Name : reservation details
    Purpose : reservation details
    Original Author : Gaurav Patel
    Created At : 03rd May 2021
*/
router.get('/reservation-details', async (req, res) => {
    try {
        req.query.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.query.time_zone = req.headers.time_zone
        }
        if (_.isEmpty(req.query) || !req.query.code || !req.query.customer_id || !req.query.reservation_id) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.query.code), null);
            return;
        }

        let response = await seekerHandler.reservationDetails(req.query);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.query.code), null);
            return;
        }
    }
});

/*
    Name : reservation cancel
    Purpose : reservation cancel
    Original Author : Gaurav Patel
    Created At : 03rd May 2021
*/
router.post('/reservation-cancel', async (req, res) => {
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        if (!req.body.code || !req.body.customer_id || !req.body.reservation_id) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.body.code), null);
            return;
        }

        let response = await seekerHandler.reservationCancel(req.body);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
            return;
        }
    }
});

/*
    Name : reservation parked
    Purpose : reservation parked
    Original Author : Gaurav Patel
    Created At : 03rd May 2021
*/
router.post('/reservation-parked', async (req, res) => {
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        if (!req.body.code || !req.body.customer_id || !req.body.reservation_id) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.body.code), null);
            return;
        }

        let response = await seekerHandler.reservationParked(req.body);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
            return;
        }
    }
});

/*
    Name : extend-parking
    Purpose : extend-parking
    Original Author : Gaurav Patel
    Created At : 03rd May 2021
*/
router.post('/extend-parking', async (req, res) => {
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        if (!req.body.code || !req.body.customer_id || !req.body.reservation_id || !req.body.to_date || !req.body.price) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.body.code), null);
            return;
        }

        let response = await seekerHandler.extendParking(req.body);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
            return;
        }
    }
});

/*
    Name : property-rating
    Purpose : property-rating
    Original Author : Gaurav Patel
    Created At : 03rd May 2021
*/
router.post('/property-rating', async (req, res) => {
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        if (!req.body.code || !req.body.customer_id || !req.body.property_id || !req.body.rating || !req.body.rating_status) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.body.code), null);
            return;
        }

        let response = await seekerHandler.propertyRating(req.body);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
            return;
        }
    }
});

/*
    Name : owner-rating
    Purpose : owner-rating
    Original Author : Gaurav Patel
    Created At : 03rd May 2021
*/
router.post('/owner-rating', async (req, res) => {
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        if (!req.body.code || !req.body.customer_id || !req.body.owner_id || !req.body.rating || !req.body.rating_status) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.body.code), null);
            return;
        }

        let response = await seekerHandler.ownerRating(req.body);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
            return;
        }
    }
});

/*
    Name : Cron For After Before Push
    Purpose : Cron For After Before Push
    Original Author : Gaurav Patel
    Created At : 29th Sep 2021
*/
router.post('/cron-for-before-after-push', async (req, res) => {
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        let response = await seekerHandler.cronBeforeAfterSendPush(req.body);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
            return;
        }
    }
});

/*
    Name : cancel-reservation
    Purpose : cancel-reservation
    Original Author : Gaurav Patel
    Created At : 19th Oct 2021
*/
router.post('/cancel-reservation-from-cron', async (req, res) => {
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        let response = await seekerHandler.cancelReservationFromCron(req.body);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
            return;
        }
    }
});

/*
    Name : complete-reservation
    Purpose : complete-reservation
    Original Author : Gaurav Patel
    Created At : 25th May 2021
*/
// router.post('/complete-reservation', async(req, res) => {
//     try {
//         req.body.time_zone = config.time_zone
//         if(req.headers.time_zone){
//             req.body.time_zone = req.headers.time_zone
//         }
//         let response = await seekerHandler.completeReservation(req.body);
//         jsonResponse(res, responseCodes.OK, errors.noError(), response);
//     } catch (error) {
//         try {
//             jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
//             return;
//         } catch (error) {
//             jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
//             return;
//         }
//     }
// });

/*
    Name : complete reservation with normal
    Purpose : complete reservation with normal
    Original Author : Gaurav Patel
    Created At : 24th Jun 2021
*/
router.post('/complete-reservation-with-app', async (req, res) => {
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        console.log("complete-reservation-with-app");

        if (!req.body.code || !req.body.customer_id || !req.body.reservation_id || !req.body.to_date) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.body.code), null);
            return;
        }
        let response = await seekerHandler.completeReservationWithApp(req.body);
        console.log("Sucess");
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.body.code), null);
            return;
        }
    }
});

/*
    Name : reservation get
    Purpose : reservation get
    Original Author : Gaurav Patel
    Created At : 1st July 2021
*/
router.get('/get-reserved-parked-list', async (req, res) => {
    try {
        req.query.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.query.time_zone = req.headers.time_zone
        }
        if (_.isEmpty(req.query) || !req.query.code || !req.query.customer_id) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.query.code), null);
            return;
        }

        let response = await seekerHandler.getReservedParkedList(req.query);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.query.code), null);
            return;
        }
    }
});

/*
    Name : near parking list
    Purpose : near parking list
    Original Author : Nikunj Prajapati
    Created At : 20 Oct 2023
*/
router.post('/near-parking-list',async(req,res)=>{
    try {
        req.body.time_zone = config.time_zone
        if (req.headers.time_zone) {
            req.body.time_zone = req.headers.time_zone
        }
        if (!req.body.latitude || !req.body.longitude) {
            jsonResponse(res, responseCodes.BadRequest, errors.missingParameters(true, req.body.latitude), null);
            return;
        }

        let response = await seekerHandler.nearParkingList(req.body);
        jsonResponse(res, responseCodes.OK, errors.noError(), response);
        
        
    } catch (error) {
        try {
            jsonResponse(res, error.code, errors.formatErrorForWire(error), null);
            return;
        } catch (error) {
            jsonResponse(res, responseCodes.InternalServer, errors.internalServer(true, req.query.code), null);
            return;
        }
    }
});

module.exports = router;